from .request_utils import post, handle_request_error

def notify_me(content="New notification!"):
    """
    send a push notification with given content 

    parameters :
    content : text content of the push notification

    returns :
    """
    res = post('notif/notifyme', 
                json={"message": content})
    if 'python_error' in res or res.status_code != 200:
        handle_request_error(res)




from datetime import datetime, timezone, timedelta
from time import sleep
from types import GeneratorType
import traceback

from .notifications import notify_me
from .request_utils import post, get, put, handle_request_error
from .account_manager import check_token_exists
from .user_token import get_token
from .cli import print_error
from datetime import datetime
from threading import Timer


END_ALERT_LOWER_BOUND = 1 ## in minutes

class EventLoop(object):
    """
    Non-blocking loop from https://stackoverflow.com/a/38317060. The method `_run` is periodically
    called with `self.interval` seconds in between
    """

    def __init__(self, interval, function, *args, **kwargs):
        self._timer = None
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.is_running = False
        self.start()

    def _run(self):
        self.is_running = False
        self.start()
        self.function(*self.args, **self.kwargs)

    def start(self):
        if not self.is_running:
            self._timer = Timer(self.interval, self._run)
            self._timer.start()
            self.is_running = True

    def stop(self):
        self._timer.cancel()
        self.is_running = False


class ntm(object):
    """
    ntm (NoTify Me) decorates an iterable and sends periodic updates.
    It is implemented as a context manager to end the job properly 
    if the  iterator is interrupted (in __exit__). 
    More info: https://docs.python.org/3/reference/datamodel.html#with-statement-context-managers

    Attributes
    ----------
    iterable : an iterable object (required)
    name : name of the job (empty by default)
    current_iteration : iteration index at which you start the job (0 by default)
    update_period : number of iteration of the loop between each update (5 by defaults)
    disable_end_message : states if you want NTM to send you a message at the end or not (false by default)
    """
    def __init__(self, iterable, name="", current_iteration=0, update_period=5, disable_end_message=False, length=None):
        """
        Creates a NTM object to track your jobs

        Parameters
        ----------
        iterable : an iterable object (required)
        name : name of the job (empty by default)
        current_iteration : iteration index at which you start the job (0 by default)
        update_period : number of iteration of the loop between each update (5 by defaults)
        disable_end_message : states if you want NTM to send you a message at the end or not (false by default)

        Returns 
        -------
        A NTM object
        
        """
        # check that a token exists, if it doesn't exist, prompt register or login
        check_token_exists()
        self.iterable = iterable
        self.name = name
        self.current_iteration = current_iteration
        if length:
            self.total_iteration = length - 1
        else:
            if isinstance(iterable, GeneratorType):
                raise RuntimeError("You need to specify the total number of iteration (`length` parameter) when using a generator")
            self.total_iteration = len(iterable)
        self.time_started = self._now()
        self.interruption = None # The server sets this attr to an error string to stop the job
        self.job_id = self._create_job()
        self.update_period = update_period
        self.already_printed_warning = False
        self.disable_end_message = disable_end_message

        ## declaration of lp as None to initialize it
        self.lp = None

    def __iter__(self):
        """
        This method is called when starting a loop. It initializes a periodic update
        and then yields element one by one, keeping track of adequate information. 
        """
        self.time_at_last_iteration = self._now()
        self.lp = EventLoop(self.update_period, self._update_job) # periodic update loop calling update_job

        for item in self.iterable:
            yield item

            self.current_iteration += 1
            self.time_at_last_iteration = self._now()
            if self.interruption:
                # Raise error when interruption is a non-empty string
                raise RuntimeError(self.interruption)
    
    def __enter__(self):
        """
        This method is called when using `with ntm(...) as ...`
        What is returned by __enter__ will be assigned to what 
        follows `as` in the `with` statement
        """
        return self

    def __exit__(self, exc_type, exc_value, tb):
        """
        This method is called when leaving the `with` statement. Closing the job
        here ensures that it is done even if an exception occurs during an iteration. 
        """
        
        if(self.lp is not None):
            # may be called before __iter__
            self.lp.stop()

        stacktrace_str = ""
        error_occurred = (exc_type is not None) or (exc_value is not None) or (tb is not None)
        if error_occurred:
            """
            If an exception occurred, traceback.format_exception will format a nice output to display
            all information about the error, including type and value.
            It returns a string array, in which each line ends by a newline.
            
            Ex output :
            Traceback (most recent call last):
              File "path/to/file/script.py", line 14, in my_function
                print(8/i)
            ZeroDivisionError: division by zero
            """
            strings = traceback.format_exception(exc_type, exc_value, tb, limit=None, chain=True)
            stacktrace_str = ''.join(strings)

        is_finished = not error_occurred
        self._job_done(is_finished, error_occurred, self._now(), stacktrace_str)

    def _date_to_str(self, date):
        return date.strftime("%Y-%m-%dT%H:%M:%SZ")

    def _now(self):
        return datetime.now(tz=timezone.utc)

    def _create_job(self):
        body = {
            "time_started_pck": self._date_to_str(self.time_started),
            "total_iteration": self.total_iteration,
            "current_iteration": self.current_iteration,
            "job_name": self.name,
        }
        res = post("job/create_job", json=body)
        if 'python_error' not in res and res.status_code == 200:
            job_id = res.json()['id']
            return job_id
        else:
            handle_request_error(res)

    def _update_job(self):
        if get_token() is None:
            # Don't update if no token registered
            # That happens when a user logs out or deletes their account while a job is running
            # No error printed because the user is warned when logging out
            return
        if not self.time_at_last_iteration:
            # if update is called before the first iteration then time_at_last is undefined
            return
        if not self.job_id:
            # print_error("Update failed due to incorrect job initializion.")
            return
        body = {
        "current_iteration": self.current_iteration,
        "time_at_last_iteration_pck": self._date_to_str(self.time_at_last_iteration)
        }
        res = put(f"job/{self.job_id}/update_job", json=body)
        if 'python_error' in res or res.status_code != 200:
            if 'python_error' not in res and 'stop' in res.json():
                #E.g. if a job has been running for too long 
                self.interruption = res.json()['failure']
            else:
                #Don't print errors repeatedly in the update loop
                if not self.already_printed_warning:
                    handle_request_error(res)
                    self.already_printed_warning = True

    def _job_done(self, is_finished, error_occurred, time_finished_pck, stacktrace_str):
        if not self.job_id:
            print_error("Job finished on an error due to incorrect job initializion.")
            notify_me("A job has finished on an exception due to incorrect senpy initialization."
                    "The final details couldn't be transmitted to the server.")
            return
        body = {
            # We need to update the current iteration to report correctly the one in which the crash happened
            "current_iteration": self.current_iteration,
            "time_finished_pck": self._date_to_str(time_finished_pck),
            "is_finished": is_finished,
            "error_occurred": error_occurred,
            "stacktrace": stacktrace_str
        }
        res = put(f"job/{self.job_id}/done", json=body)
        if 'python_error' in res or res.status_code != 200:
            notify_me("A job has finished on an exception."
                    "The final details couldn't be transmitted to the server.")
            handle_request_error(res)
        else :
            ## if we get here that means the job is done properly
            starter = ""
            if(self.name == ""):
                starter = "A job"
            else :
                starter = f"The job {self.name}"
            
            timespan = round((self._now() - self.time_started).total_seconds())
            if(not self.disable_end_message and timespan > END_ALERT_LOWER_BOUND * 60):
                ## only superieur to 1 minutes
                notify_me(starter + f" is done (in {timespan} seconds) !")




import argparse
from .account_manager import register, login, logout, change_password, delete_account
from .request_utils import print_success, print_error

def main():
    specified_register = lambda : register(on_success=lambda res : print_success('Account created!'), on_failure=print_error)
    specified_login = lambda : login(on_success=lambda res : print_success("You are logged in, welcome!"), on_failure=print_error)
    specified_change_password = lambda : change_password(on_success=lambda res : print_success('Password changed successfully.'), on_failure=print_error)
    specified_logout = lambda : logout(on_success=lambda res : print_success("You are logged out, see you!"), on_failure=print_error)
    specified_delete_account = lambda : delete_account(on_success=lambda res : print_success("Account deleted successfully."), on_failure=print_error)
    parser = argparse.ArgumentParser(prog='senpy')
    subparsers = parser.add_subparsers()

    subparsers.add_parser('register').set_defaults(func=specified_register)
    subparsers.add_parser('login').set_defaults(func=specified_login)
    subparsers.add_parser('logout').set_defaults(func=specified_logout)
    subparsers.add_parser('change_password').set_defaults(func=specified_change_password)
    subparsers.add_parser('delete_account').set_defaults(func=specified_delete_account)

    args = vars(parser.parse_args())

    if 'func' in args:
        args.pop('func')()
    else:
        parser.print_usage()
        return



import keyring

def set_token(new_token):
    """
    sets new token in the system

    parameters : 
    new_token : new value to the token 

    returns :
    """

    keyring.set_password("senpy_auth_token", "user", f"Token {new_token}")

def get_token():
    """
    reads the token and returns it

    parameters :

    returns :
    the value of the current token
    """

    return keyring.get_password("senpy_auth_token", "user")


def delete_token():
    """
    delete the stored token

    parameters :

    returns :
    """
    if get_token() is not None:
        keyring.delete_password("senpy_auth_token", "user")


from distutils.core import setup


def readme():
    with open('README.md') as f:
        return f.read()


setup(
    name="notify-me-senpy",  # name of the package on PyPi
    version="0.5.1",  # version number
    description="Simple python tracker",  # short textual description
    long_description=readme(),
    long_description_content_type='text/markdown',
    author="Notify Me SenPy",  # authors
    author_email='notify.me.senpy@gmail.com',
    url="https://github.com/magantoine/senpy-package",  # link to the repo
    keywords="PRODUCTIVITY",  # can input a list of descriptive keywords TODO COMPLETE
    license="GPLv3",  # can chose licence TODO COMPLETE
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    entry_points={
        'console_scripts': [
            'senpy = senpy.cli:main'
        ]
    },
    install_requires=[
        "colorama==0.4.4",
        "keyring==23.2.1",
        # "PyInquirer==1.0.3",
        "requests==2.26.0",
        "inquirerpy==0.3.0"
    ],
    packages=["senpy"],
    # link to the release
    # download_url="https://github.com/magantoine/senpy-package/archive/refs/tags/0.5.tar.gz"
)



### UTILS FOR REQUEST THE SERVER
import os
import sys
import colorama
from colorama import Fore, Style, AnsiToWin32
import requests
from requests.exceptions import Timeout
from .user_token import get_token

import asyncio

colorama.init(convert=True)
stream = AnsiToWin32(sys.stderr).stream
_URL = "https://api.senpy.ch/api/{}"

def create_url(tail):
    return _URL.format(tail)

def get_base_url():
    return _URL.format("")

def request_handler(tail, request_func, json=None, headers=None, add_token=True):
    if headers is None:
        headers = {"Content-Type":"application/json"}
        if(add_token):
            token = get_token()
            if token is not None:
                headers["Authorization"] = token
    
    try:
        req = request_func(create_url(tail), headers=headers, json=json, timeout=5)
        return req # Content is of type byte
    except (Timeout, Exception) as err:
        return {
            'python_error' : True, 
            'exception': str(err),
            'message':"Server unreachable, make sure you are connected to internet"
        }


# async def async_request_handler(tail, request_type, json=None, headers=None, on_success=lambda x : None, on_error=lambda x : None):
    
#     if headers is None:
#         headers = {"Content-Type":"application/json"}
#         token = get_token()
#         if token is not None:
#             headers["Authorization"] = token
#     async with aiohttp.ClientSession() as session:
#         
#         full_URL = create_url(tail)
#         func = session.get if request_type == "GET" else session.post if request_type == "POST" else session.put
#         try :
#             async with func(full_URL, headers=headers, json=json, timeout=5) as response:
#                 content = await response.json()
#                 if(content.status_code == 200):
#                     ## success we can apply the success part here
#                     on_success(content)
#                 else :
#                     ## we have an error 
#                     on_error(content)
#         except (Timeout, Exception) as err:
#             ## timeout error is triggered
#             err_arg = {
#             'python_error' : True, 
#             'exception': str(err),
#             'message':"Server unreachable, make sure you are connected to internet"
#             }
#             on_error(err_arg)
    


def get(tail, headers=None, add_token=True):
    """
    performs a get request to the server with percise root

    parameters:
    tail : the root to the query

    returns :
    the response to the request
    """
    return request_handler(tail, requests.get, headers=headers, add_token=add_token)

def post(tail, json=None, headers=None, add_token=True):
    """
    performs a post request to the server with percise root and body

    parameters:
    tail : the root to the query
    body : dict, the content to post
    json : json, the content to post

    returns :
    the response to the request
    """
    return request_handler(tail, requests.post, json=json, headers=headers, add_token=add_token)

def put(tail, json=None, headers=None, add_token=True):
    """
    performs a put request to the server with precise root and body

    parameters:
    tail : the root to the query
    body : dict, the content to post
    json : json, the content to post

    returns :
    the response to the request
    """
    return request_handler(tail, requests.put, json=json, headers=headers, add_token=add_token)

def handle_request_error(res):
    #Import here to avoid cyclic import (source: https://stackoverflow.com/a/33547682)
    from .cli import print_error
    print_error(res)

   
def print_error(res):
    """

    prints error message 

    parameters :
    res : result of the query
    """
    def print_message(error, header=None, helpers=None):
        print(Fore.RED + "Senpy - " + error.lower() + Style.RESET_ALL, file=stream)
        if(helpers is not None):
            ## help is provided

            print(Fore.LIGHTYELLOW_EX + ("Possible fix" if header is None else header) + " : ", file=stream)
            for help in helpers :
                print("\t - " + help, file=stream)
            print(Style.RESET_ALL, file=stream)


    if type(res) == str:
        print_message(res)

    elif 'python_error' in res:
        print_message(res['message'])

    elif res.status_code == 429:
        helps = ["Try again later", "You are probably trying to sending notify_me requests to frequently", "You're iterations might be too quick"]
        print_message(error="Too many requests", helpers=helps)

    elif res.status_code == 500:
        report = ["Try again later", "contact us at notify.me.senpy@gmail.com", "report the issue at https://github.com/magantoine/senpy-package"]
        print_message("An error occurred on our side, please try again later.", header="If the problem persists", helpers=report)


    else:
        data = res.json()
        if type(data) == list:
            for error in data:
                print_message(error)
        elif type(data) == dict:
            for key, errors in data.items():
                if type(errors) == list:
                    for error in errors:
                        print_message(f"{key}: {error}")
                elif type(errors) == str:
                    print_message(f"{key}: {errors}")
        else:
            print_message(str(data))

def print_success(message):
    """

    prints succes of a query

    parameters :
    message : message to print
    """
    print(Fore.GREEN + message + Style.RESET_ALL)







from __future__ import print_function, unicode_literals
from InquirerPy import prompt
from prompt_toolkit.validation import Validator, ValidationError
from .request_utils import post
import argparse
from .user_token import get_token, set_token, delete_token
from enum import Enum 
from .request_utils import print_success, print_error
from colorama import Fore, Style, AnsiToWin32
import sys

__default_func = lambda x : None

class Status(Enum):
    """
    enum for status return
    """
    SUCCESS = 0
    FAILURE = 1

class Result:
    """
    result of a query
    """
    def __init__(self, status, message):
        self.status = status ## failure or success 
        self.message = message ## message of the result

class PasswordValidator(Validator):
    def validate(self, document):
        if len(document.text) < 8:
            raise ValidationError(
                message='At least 8 characters',
                cursor_position=len(document.text))  # Move cursor to end


username =  [{
            'type': 'input',
            'message': 'Username',
            'name': 'username', #The name of the question is the request body key
        }]
password = [{
            'type': 'password',
            'message': 'Password',
            'name': 'password',
            'validate': PasswordValidator()
        }]

def register(on_success=__default_func, on_failure=__default_func):
    """
    register to the server

    parameters:
    on_success = callback called when the request succeeds
    on_failure = callback called when the requests fails
    """
    questions = username + password + [
        {
            'type': 'password',
            'message': 'Password confirmation',
            'name': 'password_confirmation',
        }]
    registered = False




    
    disclaimers ="""This is a beta version. We do not give any warranties, whether express or implied, as to the suitability or usability of the app,
its software or any of its content. We will not be liable for any loss, whether such loss is direct, indirect, special or consequential, 
suffered by any party as a result of their use of the beta app, its content and functionalities."""

    disclaimers += Fore.LIGHTYELLOW_EX + "\nDon't share any sensitive information on SenPy.\n" + Style.RESET_ALL
    ### PROMPT THE DISCLAIMER :
    print(Fore.LIGHTYELLOW_EX + "DISCLAIMERS :" + Style.RESET_ALL, file=AnsiToWin32(sys.stderr).stream)
    print(disclaimers)
    while not registered:
        data = prompt(questions)
        res = post('auth/register', data, add_token=False)
        if 'python_error' not in res and res.status_code == 200:
            token = res.json()['auth_token']
            set_token(token)
            user_id = res.json()['id']
            on_success(res)
            registered = True
            return Result(Status.SUCCESS, 'Account created')
            
        else:
            #print_error(res)
            on_failure(res)
    return Result(Status.FAILURE, res)


def login(on_success=__default_func, on_failure=__default_func):
    """
    login to the server

    parameters:
    on_success = callback called when the request succeeds
    on_failure = callback called when the requests fails
    """
    logged_in = False
    while not logged_in:
        data = prompt(username + password)
        res = post('auth/login', json=data, add_token=False)
        if 'python_error' not in res and res.status_code == 200:
            token = res.json()['auth_token']
            set_token(token)
            logged_in = True
            on_success(res)
            return Result(Status.SUCCESS, "You are logged in, welcome!")
        else:
            on_failure(res)
    
    return Result(Status.FAILURE, res)

def logout(on_success=__default_func, on_failure=__default_func):
    """
    logout from the server

    parameters:
    on_success = callback called when the request succeeds
    on_failure = callback called when the requests fails
    """
    question = {
        'type':'confirm',
        'message':"Logging out will interrupt any job updates currently running. \nDo you still want to log out?",
        'name':'confirmation'
    }
    data = prompt(question)
    if data['confirmation'] == False:
        #User doesn't want to logout anymore
        return
    res = post('auth/logout', add_token=False)
    on_success(res)
    delete_token()
    return Result(Status.SUCCESS, "You are logged out, see you!")
    
def change_password(on_success=__default_func, on_failure=__default_func):
    """
    change password in the server

    parameters:
    on_success = callback called when the request succeeds
    on_failure = callback called when the requests fails
    """
    questions = [
        {
            'type': 'password',
            'message': 'Current password',
            'name': 'current_password',
        },
        {
            'type': 'password',
            'message': 'New password',
            'name': 'new_password',
        },
        {
            'type': 'password',
            'message': 'Confirmation',
            'name': 'new_password_confirmation',
        }
    ]
    success = False
    while not success:
        data = prompt(questions)
        res = post('auth/change_password', data)
        if 'python_error' not in res and res.status_code == 200:
            on_success(res)
            success = True
            return Result(Status.SUCCESS, 'Password changed successfully.')
        else:
            on_failure(res)
    return Result(Status.FAILURE, res)

def delete_account(on_success=__default_func, on_failure=__default_func):
    """
    delete account in the server

    parameters:
    on_success = callback called when the request succeeds
    on_failure = callback called when the requests fails
    """
    question = {
        'type':'confirm',
        'message':"Deleting the account will interrupt any job updates currently running. \nDo you want to continue?",
        'name':'confirmation'
    }
    data = prompt(question)
    if data['confirmation'] == False:
        #User doesn't want to delete account anymore
        return

    logged_in = False
    while not logged_in:
        data = prompt(username + password)
        res = post('auth/delete', data)
        if 'python_error' not in res and res.status_code == 200:
            on_success(res)
            logged_in = True
            delete_token()
            return Result(Status.SUCCESS, "Account deleted successfully.")
        else:
            on_failure(res)
    return Result(Status.FAILURE, res)

def check_token_exists():
    if get_token() != None:
        return
    choices = ['Create an account', 'I already have an account']
    question = {
        'type': 'list',
        'message': 'It looks like you are not logged in a senpy account!',
        'name': 'register_or_login',
        'choices': choices,
        'default': 0
    }
    data = prompt(question)
    if data['register_or_login'] == choices[0]:
        #First choice is register
        register(on_success=lambda res : print_success('Account created!'), on_failure=print_error)
    else:
        login(on_success=lambda res : print_success("You are logged in, welcome!"), on_failure=print_error)